'''
Modified from https://github.com/lucidrains/voicebox-pytorch/blob/main/voicebox_pytorch/voicebox_pytorch.py
'''

import math
import logging
from random import random
from functools import partial
from pathlib import Path

import torch
from torch import nn, Tensor
from torch.nn import Module
import torch.nn.functional as F
from torch.cuda.amp import autocast

import torchode as to
from torchdiffeq import odeint

from beartype import beartype
from beartype.typing import Tuple, Union

from einops import rearrange, repeat, reduce, pack, unpack
from .attend import Attend

from gateloop_transformer import SimpleGateLoopLayer as GateLoop


LOGGER = logging.getLogger(__file__)

# helper functions

def exists(val):
    return val is not None

def identity(t):
    return t

def default(val, d):
    return val if exists(val) else d

def divisible_by(num, den):
    return (num % den) == 0

def is_odd(n):
    return not divisible_by(n, 2)

def coin_flip():
    return random() < 0.5

def pack_one(t, pattern):
    return pack([t], pattern)

def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]

# tensor helpers

def prob_mask_like(shape, prob, device):
    if prob == 1:
        return torch.ones(shape, device = device, dtype = torch.bool)
    elif prob == 0:
        return torch.zeros(shape, device = device, dtype = torch.bool)
    else:
        return torch.zeros(shape, device = device).float().uniform_(0, 1) < prob

def reduce_masks_with_and(*masks):
    masks = [*filter(exists, masks)]

    if len(masks) == 0:
        return None

    mask, *rest_masks = masks

    for rest_mask in rest_masks:
        mask = mask & rest_mask

    return mask

def interpolate_1d(t, length, mode = 'bilinear'):
    " pytorch does not offer interpolation 1d, so hack by converting to 2d "

    dtype = t.dtype
    t = t.float()

    implicit_one_channel = t.ndim == 2
    if implicit_one_channel:
        t = rearrange(t, 'b n -> b 1 n')

    t = rearrange(t, 'b d n -> b d n 1')
    t = F.interpolate(t, (length, 1), mode = mode)
    t = rearrange(t, 'b d n 1 -> b d n')

    if implicit_one_channel:
        t = rearrange(t, 'b 1 n -> b n')

    t = t.to(dtype)
    return t

def curtail_or_pad(t, target_length):
    length = t.shape[-2]

    if length > target_length:
        t = t[..., :target_length, :]
    elif length < target_length:
        t = F.pad(t, (0, 0, 0, target_length - length), value = 0.)

    return t

# mask construction helpers

def mask_from_start_end_indices(
    seq_len: int,
    start: Tensor,
    end: Tensor
):
    assert start.shape == end.shape
    device = start.device

    seq = torch.arange(seq_len, device = device, dtype = torch.long)
    seq = seq.reshape(*((-1,) * start.ndim), seq_len)
    seq = seq.expand(*start.shape, seq_len)

    mask = seq >= start[..., None].long()
    mask &= seq < end[..., None].long()
    return mask

def mask_from_frac_lengths(
    seq_len: int,
    frac_lengths: Tensor
):
    device = frac_lengths.device

    lengths = (frac_lengths * seq_len).long()
    max_start = seq_len - lengths

    rand = torch.zeros_like(frac_lengths, device = device).float().uniform_(0, 1)
    start = (max_start * rand).clamp(min = 0)
    end = start + lengths

    return mask_from_start_end_indices(seq_len, start, end)

# sinusoidal positions

class LearnedSinusoidalPosEmb(Module):
    """ used by @crowsonkb """

    def __init__(self, dim):
        super().__init__()
        assert divisible_by(dim, 2)
        half_dim = dim // 2
        self.weights = nn.Parameter(torch.randn(half_dim))

    def forward(self, x):
        x = rearrange(x, 'b -> b 1')
        freqs = x * rearrange(self.weights, 'd -> 1 d') * 2 * math.pi
        fouriered = torch.cat((freqs.sin(), freqs.cos()), dim = -1)
        return fouriered

# rotary positional embeddings
# https://arxiv.org/abs/2104.09864

class RotaryEmbedding(Module):
    def __init__(self, dim, theta = 50000):
        super().__init__()
        inv_freq = 1.0 / (theta ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("inv_freq", inv_freq)

    @property
    def device(self):
        return self.inv_freq.device

    @autocast(enabled = False)
    @beartype
    def forward(self, t: Union[int, Tensor]):
        if not torch.is_tensor(t):
            t = torch.arange(t, device = self.device)

        t = t.type_as(self.inv_freq)
        freqs = torch.einsum('i , j -> i j', t, self.inv_freq)
        freqs = torch.cat((freqs, freqs), dim = -1)
        return freqs

def rotate_half(x):
    x1, x2 = x.chunk(2, dim = -1)
    return torch.cat((-x2, x1), dim = -1)

@autocast(enabled = False)
def apply_rotary_pos_emb(pos, t):
    return t * pos.cos() + rotate_half(t) * pos.sin()

# convolutional positional generating module

class ConvPositionEmbed(Module):
    def __init__(
        self,
        dim,
        *,
        kernel_size,
        groups = None
    ):
        super().__init__()
        assert is_odd(kernel_size)
        groups = default(groups, dim) # full depthwise conv by default

        self.dw_conv1d = nn.Sequential(
            nn.Conv1d(dim, dim, kernel_size, groups = groups, padding = kernel_size // 2),
            nn.GELU()
        )

    def forward(self, x, mask = None):

        if exists(mask):
            mask = mask[..., None]
            x = x.masked_fill(~mask, 0.)

        x = rearrange(x, 'b n c -> b c n')
        x = self.dw_conv1d(x)
        out = rearrange(x, 'b c n -> b n c')

        if exists(mask):
            out = out.masked_fill(~mask, 0.)

        return out

# norms

class RMSNorm(Module):
    def __init__(
        self,
        dim
    ):
        super().__init__()
        self.scale = dim ** 0.5
        self.gamma = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        return F.normalize(x, dim = -1) * self.scale * self.gamma

class AdaptiveRMSNorm(Module):
    def __init__(
        self,
        dim,
        cond_dim = None
    ):
        super().__init__()
        cond_dim = default(cond_dim, dim)
        self.scale = dim ** 0.5

        self.to_gamma = nn.Linear(cond_dim, dim)
        self.to_beta = nn.Linear(cond_dim, dim)

        # init to identity

        nn.init.zeros_(self.to_gamma.weight)
        nn.init.ones_(self.to_gamma.bias)

        nn.init.zeros_(self.to_beta.weight)
        nn.init.zeros_(self.to_beta.bias)

    def forward(self, x, *, cond):
        normed = F.normalize(x, dim = -1) * self.scale

        gamma, beta = self.to_gamma(cond), self.to_beta(cond)
        gamma, beta = map(lambda t: rearrange(t, 'b d -> b 1 d'), (gamma, beta))

        return normed * gamma + beta

# attention

class MultiheadRMSNorm(Module):
    def __init__(self, dim, heads):
        super().__init__()
        self.scale = dim ** 0.5
        self.gamma = nn.Parameter(torch.ones(heads, 1, dim))

    def forward(self, x):
        return F.normalize(x, dim = -1) * self.gamma * self.scale

class Attention(Module):
    def __init__(
        self,
        dim,
        dim_head = 64,
        heads = 8,
        dropout = 0,
        flash = False,
        qk_norm = False,
        qk_norm_scale = 10
    ):
        super().__init__()
        self.heads = heads
        dim_inner = dim_head * heads

        scale = qk_norm_scale if qk_norm else None

        self.attend = Attend(dropout, flash = flash, scale = scale)

        self.qk_norm = qk_norm

        if qk_norm:
            self.q_norm = MultiheadRMSNorm(dim_head, heads = heads)
            self.k_norm = MultiheadRMSNorm(dim_head, heads = heads)

        self.to_qkv = nn.Linear(dim, dim_inner * 3, bias = False)
        self.to_out = nn.Linear(dim_inner, dim, bias = False)

    def forward(self, x, mask = None, rotary_emb = None):
        h = self.heads

        q, k, v = self.to_qkv(x).chunk(3, dim = -1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))

        if self.qk_norm:
            q = self.q_norm(q)
            k = self.k_norm(k)

        if exists(rotary_emb):
            q, k = map(lambda t: apply_rotary_pos_emb(rotary_emb, t), (q, k))

        out = self.attend(q, k, v, mask = mask)

        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)

# feedforward

class GEGLU(Module):
    def forward(self, x):
        x, gate = x.chunk(2, dim = -1)
        return F.gelu(gate) * x

def FeedForward(dim, mult = 4, dropout = 0.):
    dim_inner = int(dim * mult * 2 / 3)
    return nn.Sequential(
        nn.Linear(dim, dim_inner * 2),
        GEGLU(),
        nn.Dropout(dropout),
        nn.Linear(dim_inner, dim)
    )

# transformer

class Transformer(Module):
    def __init__(
        self,
        dim,
        *,
        depth,
        dim_head = 64,
        heads = 8,
        ff_mult = 4,
        attn_dropout = 0.,
        ff_dropout = 0.,
        num_register_tokens = 0.,
        attn_flash = False,
        adaptive_rmsnorm = False,
        adaptive_rmsnorm_cond_dim_in = None,
        use_unet_skip_connection = False,
        skip_connect_scale = None,
        attn_qk_norm = False,
        use_gateloop_layers = False,
        gateloop_use_jax = False,
    ):
        super().__init__()
        assert divisible_by(depth, 2)
        self.layers = nn.ModuleList([])

        self.rotary_emb = RotaryEmbedding(dim = dim_head)

        self.num_register_tokens = num_register_tokens
        self.has_register_tokens = num_register_tokens > 0

        if self.has_register_tokens:
            self.register_tokens = nn.Parameter(torch.randn(num_register_tokens, dim))

        if adaptive_rmsnorm:
            rmsnorm_klass = partial(AdaptiveRMSNorm, cond_dim = adaptive_rmsnorm_cond_dim_in)
        else:
            rmsnorm_klass = RMSNorm

        self.skip_connect_scale = default(skip_connect_scale, 2 ** -0.5)

        for ind in range(depth):
            layer = ind + 1
            has_skip = use_unet_skip_connection and layer > (depth // 2)

            self.layers.append(nn.ModuleList([
                nn.Linear(dim * 2, dim) if has_skip else None,
                GateLoop(dim = dim, use_jax_associative_scan = gateloop_use_jax, post_ln = True) if use_gateloop_layers else None,
                rmsnorm_klass(dim = dim),
                Attention(dim = dim, dim_head = dim_head, heads = heads, dropout = attn_dropout, flash = attn_flash, qk_norm = attn_qk_norm),
                rmsnorm_klass(dim = dim),
                FeedForward(dim = dim, mult = ff_mult, dropout = ff_dropout)
            ]))

        self.final_norm = RMSNorm(dim)

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(
        self,
        x,
        mask = None,
        adaptive_rmsnorm_cond = None
    ):
        batch, seq_len, *_ = x.shape

        # add register tokens to the left

        if self.has_register_tokens:
            register_tokens = repeat(self.register_tokens, 'n d -> b n d', b = batch)

            x, ps = pack([register_tokens, x], 'b * d')

            if exists(mask):
                mask = F.pad(mask, (self.num_register_tokens, 0), value = True)

        # keep track of skip connections

        skip_connects = []

        # rotary embeddings

        positions = seq_len

        if self.has_register_tokens:
            main_positions = torch.arange(seq_len, device = self.device, dtype = torch.long)
            register_positions = torch.full((self.num_register_tokens,), -10000, device = self.device, dtype = torch.long)
            positions = torch.cat((register_positions, main_positions))

        rotary_emb = self.rotary_emb(positions)

        # adaptive rmsnorm

        rmsnorm_kwargs = dict()
        if exists(adaptive_rmsnorm_cond):
            rmsnorm_kwargs = dict(cond = adaptive_rmsnorm_cond)

        # going through the attention layers

        for skip_combiner, maybe_gateloop, attn_prenorm, attn, ff_prenorm, ff in self.layers:

            # in the paper, they use a u-net like skip connection
            # unclear how much this helps, as no ablations or further numbers given besides a brief one-two sentence mention

            if not exists(skip_combiner):
                skip_connects.append(x)
            else:
                skip_connect = skip_connects.pop() * self.skip_connect_scale
                x = torch.cat((x, skip_connect), dim = -1)
                x = skip_combiner(x)

            if exists(maybe_gateloop):
                x = maybe_gateloop(x) + x

            attn_input = attn_prenorm(x, **rmsnorm_kwargs)
            x = attn(attn_input, mask = mask, rotary_emb = rotary_emb) + x

            ff_input = ff_prenorm(x, **rmsnorm_kwargs) 
            x = ff(ff_input) + x

        # remove the register tokens

        if self.has_register_tokens:
            _, x = unpack(x, ps, 'b * d')

        return self.final_norm(x)

# encoder decoders

class Regressor(Module):
    def __init__(
        self,
        *,
        num_cond_tokens = 500,
        dim_in = None,
        dim_out = 14,
        dim_in_proj=32,
        dim_cond_emb = 1024,
        dim = 1024,
        depth = 24,
        dim_head = 64,
        heads = 16,
        ff_mult = 4,
        ff_dropout = 0.,
        time_hidden_dim = None,
        conv_pos_embed_kernel_size = 31,
        conv_pos_embed_groups = None,
        attn_dropout = 0,
        attn_flash = False,
        attn_qk_norm = True,
        use_gateloop_layers = False,
        num_register_tokens = 16,
        p_drop_prob = 0.3, # p_drop in paper
        frac_lengths_mask: Tuple[float, float] = (0.7, 1.),
        condition_on_text = True,
        **kwargs,
    ):
        super().__init__()
        dim_in = default(dim_in, dim)
        self.dim_out=dim_out
        time_hidden_dim = default(time_hidden_dim, dim * 4)
        self.proj_in = nn.Linear(dim_out, dim_in_proj)

        self.sinu_pos_emb = nn.Sequential(
            LearnedSinusoidalPosEmb(dim),
            nn.Linear(dim, time_hidden_dim),
            nn.SiLU()
        )

        assert not (condition_on_text and not exists(num_cond_tokens)), 'number of conditioning tokens must be specified (whether phonemes or semantic token ids) if training conditional voicebox'

        if not condition_on_text:
            dim_cond_emb = 0

        self.dim_cond_emb = dim_cond_emb
        self.condition_on_text = condition_on_text
        self.num_cond_tokens = num_cond_tokens

        if condition_on_text:
            self.null_cond_id = num_cond_tokens # use last phoneme token as null token for CFG
            self.to_cond_emb = nn.Embedding(num_cond_tokens + 1, dim_cond_emb)
            
       
        
        self.p_drop_prob = p_drop_prob
        self.frac_lengths_mask = frac_lengths_mask

        self.to_embed = nn.Linear(dim_in_proj*2 + dim_cond_emb, dim)

        self.null_cond = nn.Parameter(torch.zeros(dim_in), requires_grad = False)

        self.conv_embed = ConvPositionEmbed(
            dim = dim,
            kernel_size = conv_pos_embed_kernel_size,
            groups = conv_pos_embed_groups
        )

        self.transformer = Transformer(
            dim = dim,
            depth = depth,
            dim_head = dim_head,
            heads = heads,
            ff_mult = ff_mult,
            ff_dropout = ff_dropout,
            attn_dropout= attn_dropout,
            attn_flash = attn_flash,
            attn_qk_norm = attn_qk_norm,
            num_register_tokens = num_register_tokens,
            adaptive_rmsnorm = True,
            adaptive_rmsnorm_cond_dim_in = time_hidden_dim,
            use_gateloop_layers = use_gateloop_layers
        )

        dim_out = dim_out

        self.to_pred = nn.Linear(dim, dim_out, bias = False)

    @property
    def device(self):
        return next(self.parameters()).device

    @torch.inference_mode()
    def forward_with_cond_scale(
        self,
        *args,
        cond_scale = 1.,
        **kwargs
    ):
        logits = self.forward(*args, cond_drop_prob = 0., **kwargs)

        if cond_scale == 1.:
            return logits

        null_logits = self.forward(*args, cond_drop_prob = 1., **kwargs)
        return null_logits + (logits - null_logits) * cond_scale

    def forward(
        self,
        x,
        *,
        times,
        cond_token_ids,
        self_attn_mask = None,
        cond_drop_prob = 0.1,
        target = None,
        cond = None,
        cond_mask = None,
        cond_emb = None,
    ):
        # project in, in case codebook dim is not equal to model dimensions
        
        x = self.proj_in(x)

        cond = default(cond, target)

        if exists(cond):
            cond = self.proj_in(cond)

        # shapes

        batch, seq_len, cond_dim = cond.shape
        assert cond_dim == x.shape[-1]

        # auto manage shape of times, for odeint times

        if times.ndim == 0:
            times = repeat(times, '-> b', b = cond.shape[0])

        if times.ndim == 1 and times.shape[0] == 1:
            times = repeat(times, '1 -> b', b = cond_emb.shape[0])

        # construct conditioning mask if not given

        if self.training:
            if not exists(cond_mask):
                frac_lengths = torch.zeros((batch,), device = self.device).float().uniform_(*self.frac_lengths_mask)
                cond_mask = mask_from_frac_lengths(seq_len, frac_lengths)
        else:
            if not exists(cond_mask):
                cond_mask = torch.ones((batch, seq_len), device = cond.device, dtype = torch.bool)

        cond_mask_with_pad_dim = rearrange(cond_mask, '... -> ... 1')

        # as described in section 3.2

        cond = cond * ~cond_mask_with_pad_dim

        # classifier free guidance
        if cond_emb is not None:
            cond_emb_length = cond_emb.shape[-2]
            if cond_emb_length != seq_len:
                cond_emb = rearrange(cond_emb, 'b n d -> b d n')
                cond_emb = interpolate_1d(cond_emb, seq_len)
                cond_emb = rearrange(cond_emb, 'b d n -> b n d')

                if exists(self_attn_mask):
                    self_attn_mask = interpolate_1d(self_attn_mask, seq_len)

        # concat source signal, semantic / phoneme conditioning embed, and conditioning
        # and project

        #to_concat = [*filter(exists, (x, cond_emb, cond))]
        to_concat = [*filter(exists, (x, cond_emb, cond))]
        embed = torch.cat(to_concat, dim = -1)
        x = self.to_embed(embed)

        x = self.conv_embed(x, mask = self_attn_mask) + x

        time_emb = self.sinu_pos_emb(times)

        # attend

        x = self.transformer(
            x,
            mask = self_attn_mask,
            adaptive_rmsnorm_cond = time_emb
        )

        x = self.to_pred(x)

        # if no target passed in, just return logits

        if not exists(target):
            return x

        loss_mask = reduce_masks_with_and(cond_mask, self_attn_mask)

        if not exists(loss_mask):
            return F.mse_loss(x, target)

        loss = F.l1_loss(x, target, reduction = 'none')

        loss = reduce(loss, 'b n d -> b n', 'mean')
        loss = loss.masked_fill(~loss_mask, 0.)

        # masked mean

        num = reduce(loss, 'b n -> b', 'sum')
        den = loss_mask.sum(dim = -1).clamp(min = 1e-5)
        loss = num / den

        return loss.mean()

# wrapper for the CNF

def is_probably_audio_from_shape(t):
    return exists(t) and (t.ndim == 2 or (t.ndim == 3 and t.shape[1] == 1))


    
class ConditionalFlowMatcherWrapperRegressor(Module):
    @beartype
    def __init__(
        self,
        regressor: Regressor,
        sigma = 0,
        ode_atol = 1e-5,
        ode_rtol = 1e-5,
        use_torchode = False,
        torchdiffeq_ode_method = 'midpoint',   # use midpoint for torchdiffeq, as in paper
        torchode_method_klass = to.Tsit5,      # use tsit5 for torchode, as torchode does not have midpoint (recommended by Bryan @b-chiang)
        cond_drop_prob = 0.
    ):
        super().__init__()
        self.sigma = sigma
        
        self.regressor = regressor
        self.condition_on_text = regressor.condition_on_text

        #assert not (not self.condition_on_text and exists(text_to_semantic)), 'TextToSemantic should not be passed in if not conditioning on text'
        #assert not (exists(text_to_semantic) and not exists(text_to_semantic.wav2vec)), 'the wav2vec module must exist on the TextToSemantic, if being used to condition on text'

        self.text_to_semantic = None
        

        #if self.condition_on_text and (exists(text_to_semantic) or exists(duration_predictor)):
        #    assert exists(text_to_semantic) ^ exists(duration_predictor), 'you should use either TextToSemantic from Spear-TTS, or DurationPredictor for the text / phoneme to audio alignment, but not both'

        self.cond_drop_prob = cond_drop_prob

        self.use_torchode = use_torchode
        self.torchode_method_klass = torchode_method_klass

        self.odeint_kwargs = dict(
            atol = ode_atol,
            rtol = ode_rtol,
            method = torchdiffeq_ode_method
        )

    @property
    def device(self):
        return next(self.parameters()).device

    def load(self, path, strict = True):
        # return pkg so the trainer can access it
        path = Path(path)
        assert path.exists()
        pkg = torch.load(str(path), map_location = 'cpu')
        self.load_state_dict(pkg['model'], strict = strict)
        return pkg

    @torch.inference_mode()
    def sample(
        self,
        *,
        cond_mask = None,
        steps = 3,
        cond_scale = 1.,
        cond_emb=None,
        rand_scale= 1.,
    ):
        self_attn_mask = None
        cond_token_ids = None
        cond = torch.zeros((cond_emb.shape[0], cond_emb.shape[1], self.regressor.dim_out), device = self.device)
        
        shape = cond.shape
        batch = shape[0]

        # neural ode

        self.regressor.eval()

        def fn(t, x, *, packed_shape = None):
            if exists(packed_shape):
                x = unpack_one(x, packed_shape, 'b *')

            out = self.regressor.forward_with_cond_scale(
                x,
                times = t,
                cond_token_ids = cond_token_ids,
                cond = cond,
                cond_scale = cond_scale,
                cond_mask = cond_mask,
                self_attn_mask = self_attn_mask,
                cond_emb = cond_emb,
            )

            if exists(packed_shape):
                out = rearrange(out, 'b ... -> b (...)')

            return out

        y0 = torch.randn_like(cond)*rand_scale
        if steps == 1:
            t = torch.Tensor([1.0], device = self.device)
        else:
            t = torch.linspace(0, 1, steps, device = self.device)

        if not self.use_torchode:
            LOGGER.debug('sampling with torchdiffeq')

            trajectory = odeint(fn, y0, t, **self.odeint_kwargs)
            sampled = trajectory[-1]
        else:
            LOGGER.debug('sampling with torchode')

            t = repeat(t, 'n -> b n', b = batch)
            y0, packed_shape = pack_one(y0, 'b *')

            fn = partial(fn, packed_shape = packed_shape)

            term = to.ODETerm(fn)
            step_method = self.torchode_method_klass(term = term)

            step_size_controller = to.IntegralController(
                atol = self.odeint_kwargs['atol'],
                rtol = self.odeint_kwargs['rtol'],
                term = term
            )

            solver = to.AutoDiffAdjoint(step_method, step_size_controller)
            jit_solver = torch.compile(solver)

            init_value = to.InitialValueProblem(y0 = y0, t_eval = t)

            sol = jit_solver.solve(init_value)

            sampled = sol.ys[:, -1]
            sampled = unpack_one(sampled, packed_shape, 'b *')
        return sampled

    def forward(
        self,
        x1,
        *,
        mask = None,
        cond = None,
        input_sampling_rate = None, # will assume it to be the same as the audio encoder decoder sampling rate, if not given. if given, will resample
        cond_emb = None,
    ):
        """
        following eq (5) (6) in https://arxiv.org/pdf/2306.15687.pdf
        """

        batch, seq_len, dtype, σ = *x1.shape[:2], x1.dtype, self.sigma

        # if raw audio is given, convert if audio encoder / decoder was passed in

        x0 = torch.randn_like(x1)

        # random times

        times = torch.rand((batch,), dtype = dtype, device = self.device)
        t = rearrange(times, 'b -> b 1 1')

        # sample xt (w in the paper)

        w = (1 - (1 - σ) * t) * x0 + t * x1

        flow = x1 - (1 - σ) * x0

        # predict

        self.regressor.train()
        loss = self.regressor(
            w,
            cond = cond,
            cond_mask = None,
            times = times,
            target = flow,
            self_attn_mask = mask,
            cond_token_ids = None,
            cond_drop_prob = self.cond_drop_prob,
            cond_emb = cond_emb
        )

        return loss

def expand_feature(avg_fts, durations):
    avg_fts = avg_fts[:,:,None,:]
    avg_fts = torch.cat([avg_fts, torch.zeros_like(avg_fts)],2)
    avg_fts = avg_fts.reshape(avg_fts.shape[0],-1,avg_fts.shape[-1])

    durations = durations.reshape(durations.shape[0],-1)

    expanded_fts = torch.stack([torch.repeat_interleave(avg_ft,duration,dim=0) for avg_ft,duration in zip(avg_fts, durations)])
    return expanded_fts


import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
try:
    from lightning import LightningModule
except:
    lightning = None
from transformers import HubertModel, HubertConfig, BertModel, BertConfig
from .ema_module import EMAModule
from torch.optim.lr_scheduler import LambdaLR
from torch.nn import init
from ..utils.segment_utils import get_segment, Thresholder
from ..utils.lr_schedule import COSLRLAMBDA
from ..utils.noise_utils import NoiseMixer
from huggingface_hub import hf_hub_download
import torchaudio
from pathlib import Path

    
def apply_xavier_init(m):
    if isinstance(m, (torch.nn.Linear, torch.nn.Conv1d)):
        init.xavier_uniform_(m.weight)
        if m.bias is not None:
            init.zeros_(m.bias)

class Segmenter():

    def __init__(self,
                 model_ckpt="sylber",
                 speech_upstream="facebook/hubert-base-ls960",
                 ema_decay=0.999,
                 encoding_layer = 9,
                 merge_threshold=0.8,
                 norm_threshold=2.6,
                 device='cuda',
                 **kwargs,
                ):
        super().__init__()
        self.speech_model = HubertModel(HubertConfig.from_pretrained(speech_upstream, num_hidden_layers=encoding_layer))
        
        self.enc_dim = self.speech_model.config.hidden_size
        self.encoding_layer = encoding_layer

        if model_ckpt is not None:
            if model_ckpt =="sylber":
                model_ckpt = "sylber.ckpt"
            if not Path(model_ckpt).exists():
                model_ckpt = hf_hub_download(repo_id="cheoljun95/sylber", filename=model_ckpt,)
            state_dict = torch.load(model_ckpt, map_location='cpu')
            self.speech_model.load_state_dict(state_dict, strict=False )
            print("Pre-trained checkpoint loaded")
        self.speech_model = self.speech_model.eval().to(device)
        
        if 'cuda' in device and not torch.cuda.is_available():
            print("CUDA is not available. Setting devie as CPU.")
            device = "cpu"
        self.device = device
        self.norm_threshold = norm_threshold
        self.merge_threshold=merge_threshold

    def __call__(self, wav_file, in_second=True):
        """
        Process single wav file or a list of wav files through the model
        
        Args:
            wav_file: Path to a single wav file or a list of wav file paths
            in_second: if true, segment boundaries will be in seconds. Otherwise frame indices (at 50Hz).
            
        Returns:
            For single file: Dictionary with segments and segment_features
            For multiple files: List of dictionaries, each with segments and segment_features
        """
        is_batch = isinstance(wav_file, list)
        wav_files = wav_file if is_batch else [wav_file]
        
        # Load and preprocess all wav files
        batch_wavs = []
        orig_lengths = []
        max_length = 0
        
        for file in wav_files:
            wav, sr = torchaudio.load(file)
            if sr != 16000:
                wav = torchaudio.transforms.Resample(sr, 16000)(wav)
            wav = (wav - wav.mean()) / wav.std()
            orig_lengths.append(wav.shape[1])
            batch_wavs.append(wav)
            max_length = max(max_length, wav.shape[1])
        
        # Pad wavs to the same size
        padded_wavs = []
        attention_masks = []
        
        for wav, _ in zip(batch_wavs, orig_lengths):
            padding = max_length - wav.shape[1]
            if padding > 0:
                padded_wav = torch.nn.functional.pad(wav, (0, padding))
                # Create attention mask (1 for real data, 0 for padding)
                attention_mask = torch.ones(wav.shape[1], dtype=torch.long)
                attention_mask = torch.nn.functional.pad(attention_mask, (0, padding), value=0)
            else:
                padded_wav = wav
                attention_mask = torch.ones(wav.shape[1], dtype=torch.long)
            
            padded_wavs.append(padded_wav)
            attention_masks.append(attention_mask)
        
        batch_tensor = torch.cat(padded_wavs, dim=0).to(self.device)
        attention_mask = torch.stack(attention_masks).to(self.device)
                
        with torch.no_grad():
            self.speech_model.eval()
            hidden_states = self.speech_model(batch_tensor, attention_mask=attention_mask).last_hidden_state
                    
        # Process results serially
        hidden_states = hidden_states.cpu().numpy()
        all_segments = [get_segment(states, self.norm_threshold, self.merge_threshold) for states in hidden_states]
        
        outputs = []
        for i, segments in enumerate(all_segments):
            states = hidden_states[i]
            result = {
                'segments': segments * 1.0 / 50 if in_second else segments,
                'segment_features': np.stack([states[s:e].mean(0) for s, e in segments]) if len(segments) > 0 else np.array([]),
                'hidden_states': states
            }
            outputs.append(result)
        
        return outputs if is_batch else outputs[0]
        
class Sylber(nn.Module):

    def __init__(self,
                 speech_upstream="facebook/hubert-base-ls960",
                 load_hubert=False,
                 ema_decay=0.999,
                 encoding_layer = 9,
                 do_noise_augment=False,
                 extract_target_from=None,
                 noise_mixer_configs={},
                 mask_prob=0.0,
                 segment_online = False,
                 thresholder_configs={},
                 min_mask_n = 0,
                 ema_ckpt=None,
                 merge_threshold_range=[0.5,0.7],
                 max_mask_set=1,
                 use_train_thrupdate=False,
                 **kwargs,
                ):
        super().__init__()
        if load_hubert:
            self.speech_model = HubertModel.from_pretrained(speech_upstream, num_hidden_layers=encoding_layer)
        else:
            self.speech_model = HubertModel(HubertConfig.from_pretrained(speech_upstream, num_hidden_layers=encoding_layer))
            
        self.encoding_layer = encoding_layer
        self.enc_dim = self.speech_model.config.hidden_size
        
        self.ema = None
        self.ema_decay = ema_decay
        self.ema_ckpt = ema_ckpt
        
        
        if do_noise_augment:
            self.noise_mixer = NoiseMixer(**noise_mixer_configs)
        else:
            self.noise_mixer = None
        
        
        self.segment_online = segment_online
        if segment_online:
            self.thresholder = Thresholder(**thresholder_configs)
        else:
            self.thresholder = None

        self.mask_prob = mask_prob
        self.min_mask_n = min_mask_n
        self.max_mask_set=max_mask_set

        self.merge_threshold_range=merge_threshold_range
        self.use_train_thrupdate=use_train_thrupdate
        

    def ema_step(self):
        if self.ema is None:
            self.ema = EMAModule(
                self.speech_model,
                ema_decay=self.ema_decay,
            )
            
            if self.ema_ckpt is not None:
        
                ema_dict = torch.load(self.ema_ckpt, map_location='cpu')
                self.ema.model.load_state_dict(ema_dict['ema'])
        else:
            self.ema.step(self.speech_model)
            
    def segment(self, input_values=None, features=None, attention_mask=None, mergethreshold=None, normthreshold=None, **kwargs):
        """
        """        
        if features is None:
            with torch.no_grad():
                self.speech_model.eval()
                outputs = self.speech_model(input_values, attention_mask=attention_mask)
                hidden_states = outputs.last_hidden_state
            features = hidden_states
        hidden_states = features
        if normthreshold is None:
            assert self.segment_online
            normthreshold = self.thresholder.get_threshold().item()
        with torch.no_grad():
            if mergethreshold is None:
                if self.merge_threshold_range[0]<self.merge_threshold_range[1]:
                    merge_threshold = np.random.uniform(*self.merge_threshold_range)
                else:
                    merge_threshold = self.merge_threshold_range[0]
            else:
                merge_threshold = mergethreshold
            segments=[get_segment(states.cpu().numpy(), normthreshold, merge_threshold) for states in hidden_states]

        avg_fts = []
        avg_ft_attention_masks = []
        for b in range(len(hidden_states)):
            avg_fts_ = []
            for s,e in segments[b]:
                avg_ft = features[b][s:e].mean(0)
                avg_fts_.append(avg_ft)
            if len(avg_fts_) != 0:
                avg_fts.append(torch.stack(avg_fts_))
            else:
                avg_fts.append(torch.zeros_like(features[0,:1,:]))
            avg_ft_attention_masks.append(torch.ones(len(avg_fts_), device=features.device))
            
        avg_fts = nn.utils.rnn.pad_sequence(avg_fts, batch_first=True, padding_value=0.0)
        avg_ft_attention_masks = nn.utils.rnn.pad_sequence(avg_ft_attention_masks, batch_first=True, padding_value=0)   
        
        return features, segments, avg_fts
        
        
    def forward(self,input_values, segments=None, attention_mask=None, noise=None, output_states=False,**kwargs):
        """
        """        
        with torch.no_grad():
            if self.ema is None:
                self.ema_step()
            self.ema.model.eval()
            target_hidden_states = self.ema.model(input_values, attention_mask=attention_mask).last_hidden_state

        
        if segments is None:
            assert self.segment_online
            normthreshold = self.thresholder.get_threshold().item()
            with torch.no_grad():
                norms = ((target_hidden_states**2).sum(-1)+1e-8)**.5
                if not self.use_train_thrupdate:
                    self.thresholder.update_stats(norms[norms>=normthreshold],
                                                  norms[norms<normthreshold])
                else:
                    self.thresholder.update_stats(norms[norms>=normthreshold],
                                                  None)
                norm_mask = norms>=normthreshold
                if self.merge_threshold_range[0]<self.merge_threshold_range[1]:
                    merge_threshold = np.random.uniform(*self.merge_threshold_range)
                else:
                    merge_threshold = self.merge_threshold_range[0]
                segments=[get_segment(states.cpu().numpy(), normthreshold, merge_threshold) for states in target_hidden_states]
        masks = [] 
        mask_time_indices = torch.zeros_like(target_hidden_states[...,0])
        total_mask_n = 0
        for b in range(len(target_hidden_states)):
            if len(segments[b])==0:
                continue
            mask = torch.zeros(len(segments[b]), device=target_hidden_states.device)
            mask_n = (np.random.uniform(size=len(segments[b]))<self.mask_prob).sum()
            mask_n = max(self.min_mask_n, mask_n)
            total_mask_n += mask_n
            if mask_n >0:
                mask_idxs = np.random.randint(len(segments[b]),size=mask_n)
                for mask_idx in mask_idxs:
                    mask_set= np.random.randint(1,self.max_mask_set+1)
                    mask[mask_idx:min(len(mask),mask_idx+mask_set)] = 1
                    mask_time = [segments[b][mask_idx][0],
                                 segments[b][min(len(mask),mask_idx+mask_set)-1][1]]
                    mask_time_indices[b][mask_time[0]:mask_time[1]]=1
            masks.append(mask)
        mask_time_indices = mask_time_indices>0

        if self.noise_mixer is not None:
            assert noise is not None
            with torch.no_grad():
                input_values = self.noise_mixer(input_values, noise)


        hidden_states = self.speech_model(input_values, 
                                          attention_mask=attention_mask,
                                          mask_time_indices=mask_time_indices).last_hidden_state

        if self.use_train_thrupdate:
            with torch.no_grad():
                train_norms = ((hidden_states**2).sum(-1)+1e-8)**.5
                self.thresholder.update_stats(None, train_norms[~norm_mask])
                
        averaged_target_hidden_states = torch.zeros_like(target_hidden_states)
        avg_fts = []
        trg_avg_fts = []
        avg_ft_attention_masks = []
        for b in range(len(target_hidden_states)):
            trg_avg_fts_ = []
            avg_fts_ = []
            for s,e in segments[b]:
                if np.random.uniform()<self.use_target_ft_ratio:
                    avg_ft = target_hidden_states[b][s:e].mean(0)
                else:
                    avg_ft = hidden_states[b][s:e].mean(0)
                    if self.emb_detach_ratio >0.0:
                        avg_ft = (1-self.emb_detach_ratio)*avg_ft + self.emb_detach_ratio*avg_ft.detach()
                with torch.no_grad():
                    trg_avg_ft = target_hidden_states[b][s:e].mean(0)
                averaged_target_hidden_states[b][s:e] = trg_avg_ft
                trg_avg_fts_.append(trg_avg_ft)
                avg_fts_.append(avg_ft)
            if len(avg_fts_)==0:
                continue
            trg_avg_fts.append(torch.stack(trg_avg_fts_))
            avg_fts.append(torch.stack(avg_fts_))
            avg_ft_attention_masks.append(torch.ones(len(trg_avg_fts_), device=hidden_states.device))

        distillation_loss = ((hidden_states-averaged_target_hidden_states)**2).sum(-1).mean()
        
        outputs = {'distillation_loss': distillation_loss,}


        if output_states:
            outputs['avg_fts'] =avg_fts
            outputs['avg_ft_attention_masks'] = avg_ft_attention_masks
            outputs['speech_states'] = hidden_states
            outputs['teacher_speech_states'] = averaged_target_hidden_states
            outputs['lm_logits']=lm_outputs
            outputs['teacher_lm_logits'] = teacher_lm_outputs
            outputs['segments'] = segments

        return outputs

class SylberTrainer(LightningModule):

    def __init__(self, loss_coefs,lr=0.0001, warmup_steps=5000, total_steps=500000,
                 min_factor=1.0,hold_steps=0, accumulate_grad_batches=1,
                 **model_configs):
        super().__init__()
        self.loss_coefs = loss_coefs
        
        self.net = Sylber(**model_configs).to(torch.float)
        self.lr = lr
        self.total_steps =total_steps
        self.warmup_steps = warmup_steps
        self.min_factor = min_factor
        self.hold_steps = hold_steps
        self.accumulate_grad_batches = accumulate_grad_batches
        
    def forward(self, **kwargs):
        return self.net(**kwargs)
        
    def training_step(self, batch, batch_idx):
        
        if (self.global_step%self.accumulate_grad_batches) ==0:
            self.net.ema_step()
        outputs = self.net(**batch)
        
        loss_val = 0
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'train_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'train_loss', loss_val)
        if self.net.thresholder is not None:
            self.log('normthreshold', self.net.thresholder.get_threshold().item())
        return loss_val

        
    def validation_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        loss_val = 0
        
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'val_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'val_loss', loss_val,sync_dist=True)
            
        return loss_val

    def configure_optimizers(self):
        
        opt_fun = torch.optim.AdamW
        opt = opt_fun(self.net.parameters(),lr=self.lr,eps=1e-4, betas=(0.9, 0.95), weight_decay = 0.1)
        lr_lambda = COSLRLAMBDA(self.warmup_steps, self.total_steps, self.min_factor,  self.hold_steps)
        sch = LambdaLR(opt, lr_lambda)
        return [opt],[{"scheduler": sch, "interval": "step"}]



#!/usr/bin/env python3

"""
Modified from fairseq repo 
https://github.com/facebookresearch/fairseq/blob/main/fairseq/modules/ema_module.py
Used for EMA tracking a given pytorch module. The user is responsible for calling step()
and setting the appropriate decay
"""

import copy
import logging

import torch

multi_tensor_l2norm_available = False
logger = logging.getLogger(__name__)

'''
for documentation.
all arguments are copied to EMAModule
@dataclass
class EMAModuleConfig(FairseqDataclass):
    ema_decay: float = field(
        default=0.9999, metadata={"help": "decay for exponential moving average model"}
    )
    ema_fp32: bool = field(
        default=False,
        metadata={"help": "If true, store EMA model in fp32 even if model is in fp16"},
    )
    add_missing_params: bool = True
    log_norms: bool = False
'''

class EMAModule:
    """Exponential Moving Average of Fairseq Models"""

    def __init__(
        self,
        model,
        ema_decay = 0.9999,
        ema_fp32 = False,
        add_missing_params=True,
        log_norms = False,
        copy_model=True,
        device=None,
        skip_keys=None,
    ):
        """
        @param model model to initialize the EMA with
        @param config EMAConfig object with configuration like
        ema_decay, ema_update_freq, ema_fp32
        @param device If provided, copy EMA to this device (e.g. gpu).
        Otherwise EMA is in the same device as the model.
        """


        if copy_model:
            self.model = copy.deepcopy(model)
            self.model.requires_grad_(False)
        else:
            self.model = model

        self.decay = ema_decay
        self.skip_keys = skip_keys or set()
        self.add_missing_params = add_missing_params
        self.fp32_params = {}
        self.ema_fp32=ema_fp32

        if device is not None:
            logging.info(f"Copying EMA model to device {device}")
            self.model = self.model.to(device=device)

        if ema_fp32:
            self.build_fp32_params()

        self.log_norms = log_norms and multi_tensor_l2norm_available
        self.logs = {}

    def build_fp32_params(self, state_dict=None):
        """
        Store a copy of the EMA params in fp32.
        If state dict is passed, the EMA params is copied from
        the provided state dict. Otherwise, it is copied from the
        current EMA model parameters.
        """
        if not self.ema_fp32:
            raise RuntimeError(
                "build_fp32_params should not be called if ema_fp32=False. "
                "Use ema_fp32=True if this is really intended."
            )

        if state_dict is None:
            state_dict = self.model.state_dict()

        def _to_float(t):
            return t.float() if torch.is_floating_point(t) else t

        for param_key in state_dict:
            if param_key in self.fp32_params:
                if param_key == "__sq_mom":
                    self.fp32_params[param_key] = state_dict[param_key]
                else:
                    self.fp32_params[param_key].copy_(state_dict[param_key])
            else:
                self.fp32_params[param_key] = _to_float(state_dict[param_key])
                if "__sq_mom" in self.fp32_params:
                    self.fp32_params["__sq_mom"][param_key] = torch.zeros_like(
                        self.fp32_params[param_key]
                    )

    def restore(self, state_dict, build_fp32_params=False):
        """Load data from a model spec into EMA model"""
        self.model.load_state_dict(state_dict, strict=False)
        if build_fp32_params:
            self.build_fp32_params(state_dict)

    def set_decay(self, decay, weight_decay=None):
        self.decay = decay
        if weight_decay is not None:
            self.weight_decay = weight_decay

    def get_decay(self):
        return self.decay

    def _step_internal(self, new_model):
        """One update of the EMA model based on new model weights"""
        decay = self.decay

        ema_state_dict = {}
        ema_params = (
            self.fp32_params if self.ema_fp32 else self.model.state_dict()
        )

        new_p = []
        ema_p = []

        for key, param in new_model.named_parameters():
            if isinstance(param, dict):
                continue

            if not self.add_missing_params and key not in ema_params:
                continue

            try:
                ema_param = ema_params[key]
            except KeyError:
                ema_param = (
                    param.float().clone() if param.ndim == 1 else copy.deepcopy(param)
                )
                ema_params[key] = ema_param

            if param.shape != ema_param.shape:
                raise ValueError(
                    "incompatible tensor shapes between model param and ema param"
                    + "{} vs. {}".format(param.shape, ema_param.shape)
                )

            if "version" in key:
                # Do not decay a model.version pytorch param
                continue

            lr = 1 - decay

            if key in self.skip_keys or not param.requires_grad:
                ema_params[key].copy_(param.to(dtype=ema_param.dtype).data)
                ema_param = ema_params[key]
            else:
                if self.log_norms:
                    new_p.append(param)
                    ema_p.append(ema_param)

                ema_param.mul_(1 - lr)
                ema_param.add_(param.data.to(dtype=ema_param.dtype), alpha=lr)

            ema_state_dict[key] = ema_param

        for key, param in new_model.named_buffers():
            ema_state_dict[key] = param

        if self.log_norms:
            if "model_norm" in self.logs:
                self.prev_model_norm = self.logs["model_norm"]

            chunk_size = 2048 * 32
            has_inf = torch.zeros(
                (1, 1), dtype=torch.int, device=next(new_model.parameters()).device
            )

            new_norm = multi_tensor_l2norm(chunk_size, has_inf, [new_p], False)
            old_norm = multi_tensor_l2norm(chunk_size, has_inf, [ema_p], False)

            self.logs["model_norm"] = new_norm[0]
            self.logs["ema_norm"] = old_norm[0]

        self.restore(ema_state_dict, build_fp32_params=False)

    @torch.no_grad()
    def step(self, new_model):
        self._step_internal(new_model)

    def reverse(self, model):
        """
        Load the model parameters from EMA model.
        Useful for inference or fine-tuning from the EMA model.
        """
        d = self.model.state_dict()
        if "_ema" in d:
            del d["_ema"]

        model.load_state_dict(d, strict=False)
        return model



import torch
import torch.nn as nn
from vector_quantize_pytorch import GroupedResidualVQ
import numpy as np

def FeedForward(dim, dropout = 0.):
    dim_inner = dim
    return nn.Sequential(
        nn.Linear(dim, dim_inner),
        nn.ReLU(inplace=True),
        nn.Dropout(dropout),
        nn.Linear(dim_inner, dim)
    )

class FFEncoder(nn.Module):
    def __init__(self,input_dim, output_dim, hidden_dims, dropout=0.0,):
        super().__init__()
        self.input_dim = input_dim
        modules = []
        
        for dim in hidden_dims:
            modules.append(nn.Linear(input_dim,dim))
            modules.append(FeedForward(dim, dropout=dropout))
            input_dim = dim
        
        modules.append(nn.Linear(input_dim,output_dim))
        
        self.mlp = nn.Sequential(*modules)
        
    def forward(self, x):
        return self.mlp(x)

def _unit_norm(x):
    norm = (((x**2).sum(-1)+1e-5)**.5)[...,None]
    norm[norm==0] = 1.0
    x = x/norm
    return x

def _unit_norm_sep(x, spnorm, offset):
    if spnorm:
        x = torch.cat([_unit_norm(x[...,:-offset]), _unit_norm(x[...,-offset:])],-1)
        return x
    else:
        return _unit_norm(x)


def load_quantizer(config=None, ckpt=None):
    if config !=None:
        if not isinstance(config, dict):
            if not isinstance(config, dict):
                if config[-5:] == '.ckpt':
                    return load_quantizer(config=None, ckpt=config)
                else:
                    import yaml
                    with open(config) as f:
                        config = yaml.load(f, Loader=yaml.Loader)
        if 'model' in config.keys():
            config = config['model']
        state_dict = None
    else:
        assert ckpt != None
        ckpt = torch.load(ckpt)
        config = ckpt['config']
        state_dict = ckpt['state_dict']
        
    quantizer = Quantizer(**config)
    if ckpt != None and state_dict==None:
        ckpt = torch.load(ckpt)
        if 'state_dict' in ckpt.keys():
            state_dict = ckpt['state_dict']
        else:
            state_dict = ckpt
            
    if state_dict != None:
        quantizer.load_state_dict(state_dict, strict=True)
    quantizer = quantizer.eval()
    return quantizer

def load_km_quantizer(centroids, normalize=False):
    return KMQuantizer(centroids, normalize=normalize).eval()

def load_residualkm_quantizer(centroids, centroids2, normalize=False):
    return ResidualKMQuantizer(centroids, centroids2).eval()


class KMQuantizer(nn.Module):
    def __init__(self, centroids,normalize=False ):
        super().__init__()
        centroids = np.load(centroids)
        
        self.vq = GroupedResidualVQ(dim=768,
                                   num_quantizers = 1,
                                   codebook_size=centroids.shape[-1],
                                   decay=1.0,
                                    quantize_dropout=False,
                                    kmeans_init=False
                                   ) 
        device = self.vq.rvqs[0].layers[0]._codebook.embed
        self.vq.rvqs[0].layers[0]._codebook.embed = torch.from_numpy(centroids).to(device)[None,...]
        self.vq.rvqs[0].layers[0]._codebook.initted = torch.Tensor([True])
        self.normalize = normalize
    

    def get_indices(self, token):
        was_training = self.training
        self = self.eval()
        with torch.no_grad():
            if self.normalize:
                token =token/(((token**2).sum(-1)+1e-8)**.5)[...,None]*6
            outputs = self(token)
        
        if was_training:
            self = self.train()
            
        return outputs['indices']
            
    def _get_device(self):
        return self.encoder.mlp[0].weight.device
        
    def forward(self, token):
        # token: (B, L, d)
        quantized, indices, vq_loss = self.vq(token)
        
        outputs = {"indices":indices[0], "quantize":quantized, "non_quantized":token,
                   "commitment_loss": vq_loss}

        return outputs

    def decode(self, indices):
        # indices: (B, L, # art codebooks + # pitch codebooks)
        indices = indices.clip(0)
        indices = indices[...,:1]
        
        quantized = self.vq.get_output_from_indices(indices.unsqueeze(0))
        return quantized

class ResidualKMQuantizer(nn.Module):
    def __init__(self, centroids,centroids2,normalize=False ):
        super().__init__()
        self.km = KMQuantizer(centroids)
        self.km2 = KMQuantizer(centroids2)
    

    def get_indices(self, token):
        was_training = self.training
        self = self.eval()
        with torch.no_grad():
            idxs1 = self.km.get_indices(token)
            z_q = self.km.decode(idxs1)
            idxs2 = self.km2.get_indices(token-z_q)
            #outputs = self(token)
        
        if was_training:
            self = self.train()
        return torch.cat([idxs1, idxs2],-1)
            
    def _get_device(self):
        return self.encoder.mlp[0].weight.device
        
    def forward(self, token):
        # token: (B, L, d)
        
        quantized, indices, vq_loss = self.vq(token)
        
        outputs = {"indices":indices[0], "quantize":quantized, "non_quantized":token,
                   "commitment_loss": vq_loss}

        return outputs

    def decode(self, indices):
        # indices: (B, L, # art codebooks + # pitch codebooks)
        z_q1 = self.km.decode(indices[...,:1])
        z_q2 = self.km2.decode(indices[...,1:])
        quantized = z_q1+z_q2
        return quantized
        

        



class Quantizer(nn.Module):
    def __init__(self, encoder_configs, art_vq_configs, pitch_vq_configs,
                 unit_norm_encoder_input=True, unit_norm_encoder_output=True,
                 keep_blank_zero=True, pitch_emb_dim=8, separate_norm=True):
        super().__init__()
        self.encoder = FFEncoder(**encoder_configs)
        self.art_vq = GroupedResidualVQ(**art_vq_configs) 
        self.pitch_vq = GroupedResidualVQ(**pitch_vq_configs) 
        self.pitch_emb_dim = pitch_emb_dim
        self.art_emb_dim = encoder_configs['output_dim'] - self.pitch_emb_dim
        self.keep_blank_zero = keep_blank_zero
        self.separate_norm = separate_norm
        self.unit_norm_encoder_input = unit_norm_encoder_input
        self.unit_norm_encoder_output = unit_norm_encoder_output
        self.art_codebook_num = art_vq_configs["num_quantizers"]
        self.pitch_codebook_num = pitch_vq_configs["num_quantizers"]

    def get_indices(self, token):
        was_training = self.training
        self = self.eval()
        with torch.no_grad():
            outputs = self(token)

        if was_training:
            self = self.train()
            
        return outputs['indices']
            
    def _get_device(self):
        return self.encoder.mlp[0].weight.device
        
    def forward(self, token):
        # token: (B, L, d)
        
        non_blank_mask = (token**2).sum(-1)>0
        
        if self.unit_norm_encoder_input:
            token = _unit_norm(token)
           
        token = self.encoder(token)
        if self.unit_norm_encoder_output:
            token =_unit_norm_sep(token, self.separate_norm, self.pitch_emb_dim)
            
        if self.keep_blank_zero:
            token[~non_blank_mask] = 0.0
    
        art_token, pitch_token = token[...,:-self.pitch_emb_dim],  token[...,-self.pitch_emb_dim:]
        art_quantized, art_indices, art_vq_loss = self.art_vq(art_token)
        pitch_quantized, pitch_indices, pitch_vq_loss = self.pitch_vq(pitch_token)
        
        quantized = torch.cat([art_quantized, pitch_quantized],-1)
        
        if self.unit_norm_encoder_output:
            quantized = _unit_norm_sep(quantized, self.separate_norm, self.pitch_emb_dim)

        indices = torch.cat([art_indices[0], pitch_indices[0]], -1)  # (B, L, # art codebooks + # pitch codebooks)
        outputs = {"indices":indices, "quantize":quantized, "non_quantized":token,
                   "commitment_loss": art_vq_loss+pitch_vq_loss}

        return outputs

    def decode(self, indices):
        # indices: (B, L, # art codebooks + # pitch codebooks)
        indices = indices.clip(0)
        art_indices = indices[...,:self.art_codebook_num]
        pitch_indices = indices[...,self.art_codebook_num:]
        
        art_quantized = self.art_vq.get_output_from_indices(art_indices.unsqueeze(0))
        pitch_quantized = self.pitch_vq.get_output_from_indices(pitch_indices.unsqueeze(0))
        
        quantized = torch.cat([art_quantized, pitch_quantized],-1)
        
        if self.unit_norm_encoder_output:
            quantized = _unit_norm_sep(quantized, self.separate_norm, self.pitch_emb_dim)

        return quantized

        



from setuptools import setup, find_packages

setup(
    name='sylber',  
    version='0.1.1',
    author='Cheol Jun Cho, Nicholas Lee, Akshat Gupta, Dhruv Agarwal, Ethan Chen, Alan Black, Gopala Anumanchipalli',
    author_email='cheoljun@berkeley.edu',
    description='Python code for "Sylber: Syllabic Embedding Representation of Speech from Raw Audio"',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Berkeley-Speech-Group/sylber',  
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',  
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
    install_requires=[
        'numpy',
        'soundfile',
        'librosa',
        'torch',
        'torchaudio',
        'transformers',
        'huggingface-hub',
        'tqdm',
        'speech-articulatory-coding',
        'typeguard>=4.0.1',
        'torchode==1.0.0',
        'torchdiffeq==0.2.4',
        'beartype==0.19.0',
        'gateloop-transformer==0.2.5',
        'vector-quantize-pytorch==1.18.5',
    ],
    include_package_data=False,  
    license='Apache-2.0',  
)



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
try:
    from lightning import LightningModule
except:
    lightning = None
from transformers import HubertModel, HubertConfig
from torch.optim.lr_scheduler import LambdaLR
from .flowmatching import Regressor, ConditionalFlowMatcherWrapperRegressor
from ..utils.segment_utils import get_segment, Thresholder
from ..utils.lr_schedule import COSLRLAMBDA
from .quantizer import load_km_quantizer


class RFF(nn.Module):
    def __init__(self, dim, dropout=0.05):
        super().__init__()
        self.linear1 = nn.Linear(dim, dim)
        self.linear2 = nn.Linear(dim, dim)
        self.dropout1= nn.Dropout(dropout)
        self.dropout2= nn.Dropout(dropout)
        self.norm = nn.LayerNorm(dim)
        self.activation = nn.ReLU()
        
    def forward(self, x):
        
        x2 = self.linear2(self.dropout1(self.activation(self.linear1(x))))
        x = x + self.dropout2(x2)
        x = self.norm(x)
        return x


class MLP(nn.Module):
    def __init__(self,input_dim, output_dim, hidden_dims, dropout=0.05):
        super().__init__()
        #self.input_layer = nn.Linear(input_dim, hidden_dims[0])
        #self.output_layer = nn.Linear(hidden_dims[-1],output_dim)
        
        modules = []
        
        for dim in hidden_dims:
            modules.append(nn.Linear(input_dim,dim))
            modules.append(RFF(dim,dropout))
            input_dim = dim
        
        modules.append(nn.Linear(hidden_dims[-1],output_dim))
        
        self.mlp = nn.Sequential(*modules)
        
    def forward(self, x):
        return self.mlp(x)
        

class SegmentSynthesis(nn.Module):

    def __init__(self,
                 speech_upstream="facebook/hubert-base-ls960",
                 encoding_layer = 9,
                 regressor_configs={},
                 noise_mixer_configs={},
                 segment_online = False,
                 thresholder_configs={},
                 input_configs={},
                 merge_threshold_range=[0.8,0.8],
                 quantizer=None,
                 pitch_amp=5,
                 art_input_configs={},
                 **kwargs,
                ):
        super().__init__()
        self.speech_model = HubertModel(HubertConfig.from_pretrained(speech_upstream, num_hidden_layers=encoding_layer))
        self.speech_model.requires_grad_(False)
        self.enc_dim = self.speech_model.config.hidden_size
        self.input_model = MLP(self.enc_dim, **input_configs)
        self.input_dim = input_configs['output_dim']
        self.encoding_layer = encoding_layer
        
        self.regressor = Regressor(**regressor_configs)
        self.cfm_wrapper = ConditionalFlowMatcherWrapperRegressor(
                regressor = self.regressor,
                sigma=regressor_configs['sigma'],
            )
       
        self.segment_online = segment_online
        if segment_online:
            self.thresholder = Thresholder(**thresholder_configs)
        else:
            self.thresholder = None
        
        self.merge_threshold_range=merge_threshold_range
        if quantizer is not None:
            if residual_quantizer is not None:
                self.quantizer = load_km_quantizer(quantizer,residual_quantizer)
            else:
                self.quantizer = load_km_quantizer(quantizer,normalize=normalize_embed)
            self.quantizer.requires_grad_(False)
        else:
            self.quantizer = None
        self.pitch_amp = pitch_amp
            
    def resynthesize(self,input_values=None,attention_mask=None,features=None,  steps=5,rand_scale=0.0,merge_threshold=0.8,normthreshold=None,
                    prosody_steps=None, prosody_rand_scale=None):
        """
        """      
        if features is None:
            with torch.no_grad():
                self.speech_model.eval()
                target_hidden_states = self.speech_model(input_values, attention_mask=attention_mask).last_hidden_state
            if normthreshold is None:
                normthreshold = self.thresholder.get_threshold().item()
            with torch.no_grad():
                norms = ((target_hidden_states**2).sum(-1)+1e-8)**.5
                
                segments=[get_segment(states.cpu().numpy(), normthreshold, merge_threshold) for states in target_hidden_states]
    
            valid_segments = [segment for segment in segments if len(segment) >0]
            averaged_target_hidden_states = torch.zeros_like(target_hidden_states[:,:,:])
            trg_avg_fts = []
            for b in range(len(target_hidden_states)):
                trg_avg_fts_ = []
                for s,e in segments[b]:
                    with torch.no_grad():
                        trg_avg_ft = target_hidden_states[b][s:e].mean(0)
                        if self.quantizer is not None:
                            with torch.no_grad():
                                q_idxs = self.quantizer.get_indices(trg_avg_ft.unsqueeze(0))
                                q_decoded = self.quantizer.vq.get_output_from_indices(q_idxs.unsqueeze(0)) #/6*trg_avg_ft_norm
                                trg_avg_ft = q_decoded[0]
                        trg_avg_ft = trg_avg_ft
                    averaged_target_hidden_states[b][s:e] = trg_avg_ft
                    trg_avg_fts_.append(trg_avg_ft)
                if len(trg_avg_fts_) != 0:
                    trg_avg_fts.append(torch.stack(trg_avg_fts_))
        else:
            averaged_target_hidden_states = features
            norms = ((features**2).sum(-1))**.5
            normthreshold = 1e-4
            segments = None
        input = self.input_model(averaged_target_hidden_states)
        input[norms<normthreshold]=0.0
        art = self.cfm_wrapper.sample(cond_emb=input,steps=steps,rand_scale=rand_scale)        
        art = art.clone()
        art[...,12] = art[...,12]/self.pitch_amp
        return art,segments
        
    def forward(self,input_values=None, features=None, segments=None, art=None, attention_mask=None, noise=None, **kwargs):
        """
        """        
        if features is None:
            if self.noise_mixer is not None and noise is not None:
                with torch.no_grad():
                    input_values = self.noise_mixer(input_values, noise)
                    
            with torch.no_grad():
                self.speech_model.eval()
                target_hidden_states = self.speech_model(input_values, attention_mask=attention_mask).last_hidden_state
    
            if segments is None:
                assert self.segment_online
                normthreshold = self.thresholder.get_threshold().item()
                with torch.no_grad():
                    norms = ((target_hidden_states**2).sum(-1)+1e-8)**.5
                    if self.merge_threshold_range[0]<self.merge_threshold_range[1]:
                        merge_threshold = np.random.uniform(*self.merge_threshold_range)
                    else:
                        merge_threshold = self.merge_threshold_range[0]
                    segments=[get_segment(states.cpu().numpy(), normthreshold, merge_threshold) for states in target_hidden_states]
    
            
            valid_segments = [segment for segment in segments if len(segment) >0]
            averaged_target_hidden_states = torch.zeros_like(target_hidden_states[:,:,:])
            trg_avg_fts = []
            for b in range(len(target_hidden_states)):
                trg_avg_fts_ = []
                for s,e in segments[b]:
                    trg_avg_ft = target_hidden_states[b][s:e].mean(0)
                    trg_avg_ft = trg_avg_ft
                    averaged_target_hidden_states[b][s:e] = trg_avg_ft
                    trg_avg_fts_.append(trg_avg_ft)
                if len(trg_avg_fts_) != 0:
                    trg_avg_fts.append(torch.stack(trg_avg_fts_))
            input = self.input_model(averaged_target_hidden_states)
            input[norms<normthreshold]=0.0
            features = averaged_target_hidden_states
        else:
            norms = ((features**2).sum(-1))**.5
            if self.quantizer is not None:
                with torch.no_grad():
                    q_idxs = self.quantizer.get_indices(features)
                    features = self.quantizer.decode(q_idxs)
                    features[norms<1e-4] = 0.0
            input = self.input_model(features)
            normthreshold=1e-4
            input[norms<normthreshold]=0.0
        
        art[...,12] *= self.pitch_amp
        outputs = {'cfm_loss': cfm_loss}
        
        return outputs

class SegmentSynthesisTrainer(LightningModule):

    def __init__(self, loss_coefs,lr=0.0001, warmup_steps=5000, total_steps=500000, min_factor=0.05, 
                 **model_configs):
        super().__init__()
        self.loss_coefs = loss_coefs
        
        self.net = SegmentSynthesis(**model_configs).to(torch.float)
        self.lr = lr
        self.total_steps =total_steps
        self.warmup_steps = warmup_steps
        self.min_factor = min_factor
        
    def forward(self, **kwargs):
        return self.net(**kwargs)
    
    def training_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        
        loss_val = 0
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'train_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'train_loss', loss_val)
        
        return loss_val

        
    def validation_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        loss_val = 0
        
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'val_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'val_loss', loss_val,sync_dist=True)
            
        return loss_val

    def configure_optimizers(self):
        
        opt_fun = torch.optim.AdamW
        opt = opt_fun(self.net.parameters(),lr=self.lr,eps=1e-4, betas=(0.9, 0.95), weight_decay = 0.1)
        lr_lambda = COSLRLAMBDA(self.warmup_steps, self.total_steps, self.min_factor)
        sch = LambdaLR(opt, lr_lambda)
        return [opt],[{"scheduler": sch, "interval": "step"}]
